import os
import pathlib
import re


changelogVersionPattern = re.compile(r"## \[(.+)\]")
githubRefPattern = re.compile(r"refs/tags/v(.*)")


changelogPath = pathlib.Path(__file__).resolve().parent.parent.parent / "CHANGELOG.md"
changelog = changelogPath.read_text("utf-8")

m = githubRefPattern.match(os.getenv("GITHUB_REF"))
version = m.group(1)

notes = []

collecting = False
for line in changelog.splitlines():
    m = changelogVersionPattern.match(line)
    if m is not None:
        if collecting:
            break
        elif m.group(1) == version:
            collecting = True
    elif collecting:
        notes.append(line)

print("\n".join(notes).strip().replace("\n", "%0A"))



import logging
import math
import skia
from fontTools.misc.transform import Transform
from fontTools.pens.basePen import BasePen
from fontTools.pens.pointPen import SegmentToPointPen
from .gstate import TextStyle


# TODO:
# - textBox
# MAYBE:
# - contours
# - expandStroke
# - intersectionPoints
# - offCurvePoints
# - onCurvePoints
# - optimizePath
# - points
# - svgClass
# - svgID
# - svgLink
# - traceImage


class BezierPath(BasePen):

    def __init__(self, path=None, glyphSet=None):
        super().__init__(glyphSet)
        if path is None:
            path = skia.Path()
        self.path = path

    def _moveTo(self, pt):
        self.path.moveTo(*pt)

    def _lineTo(self, pt):
        self.path.lineTo(*pt)

    def _curveToOne(self, pt1, pt2, pt3):
        x1, y1 = pt1
        x2, y2 = pt2
        x3, y3 = pt3
        self.path.cubicTo(x1, y1, x2, y2, x3, y3)

    def _qCurveToOne(self, pt1, pt2):
        x1, y1 = pt1
        x2, y2 = pt2
        self.path.quadTo(x1, y1, x2, y2)

    def _closePath(self):
        self.path.close()

    def _endPath(self):
        pass

    def arc(self, center, radius, startAngle, endAngle, clockwise):
        cx, cy = center
        diameter = radius * 2
        rect = (cx - radius, cy - radius, diameter, diameter)
        sweepAngle = (endAngle - startAngle) % 360
        if clockwise:
            sweepAngle -= 360
        self.path.arcTo(rect, startAngle, sweepAngle, False)

    def arcTo(self, point1, point2, radius):
        self.path.arcTo(point1, point2, radius)

    def rect(self, x, y, w, h):
        self.path.addRect((x, y, w, h))

    def oval(self, x, y, w, h):
        self.path.addOval((x, y, w, h))

    def line(self, pt1, pt2):
        points = [(x, y) for x, y in [pt1, pt2]]
        self.path.addPoly(points, False)

    def polygon(self, firstPoint, *points, close=True):
        points = [(x, y) for x, y in (firstPoint,) + points]
        self.path.addPoly(points, close)

    def pointInside(self, point):
        x, y = point
        return self.path.contains(x, y)

    def bounds(self):
        if self.path.countVerbs() == 0:
            return None
        return tuple(self.path.computeTightBounds())

    def controlPointBounds(self):
        if self.path.countVerbs() == 0:
            return None
        return tuple(self.path.getBounds())

    def reverse(self):
        path = skia.Path()
        path.reverseAddPath(self.path)
        self.path = path

    def appendPath(self, other):
        self.path.addPath(other.path)

    def copy(self):
        path = skia.Path(self.path)
        return BezierPath(path=path)

    def translate(self, x, y):
        self.path.offset(x, y)

    def scale(self, x, y=None, center=(0, 0)):
        if y is None:
            y = x
        self.transform((x, 0, 0, y, 0, 0), center=center)

    def rotate(self, angle, center=(0, 0)):
        t = Transform()
        t = t.rotate(math.radians(angle))
        self.transform(t, center=center)

    def skew(self, x, y=0, center=(0, 0)):
        t = Transform()
        t = t.skew(math.radians(x), math.radians(y))
        self.transform(t, center=center)

    def transform(self, transform, center=(0, 0)):
        cx, cy = center
        t = Transform()
        t = t.translate(cx, cy)
        t = t.transform(transform)
        t = t.translate(-cx, -cy)
        matrix = skia.Matrix()
        matrix.setAffine(t)
        self.path.transform(matrix)

    def drawToPen(self, pen):
        it = skia.Path.Iter(self.path, False)
        needEndPath = False
        for verb, points in it:
            penVerb, startIndex, numPoints = _pathVerbsToPenMethod.get(verb, (None, None, None))
            if penVerb is None:
                continue
            assert len(points) == numPoints, (verb, numPoints, len(points))
            if penVerb == "conicTo":
                # We should only call _convertConicToCubicDirty()
                # if it.conicWeight() == sqrt(2)/2, but skia-python doesn't
                # give the correct value.
                # https://github.com/kyamagu/skia-python/issues/116
                # if abs(it.conicWeight() - 0.707...) > 1e-10:
                #     logging.warning("unsupported conic form (weight != sqrt(2)/2): conic to cubic conversion will be bad")
                # TODO: we should fall back to skia.Path.ConvertConicToQuads(),
                # but that call is currently also not working.
                pen.curveTo(*_convertConicToCubicDirty(*points))
            elif penVerb == "closePath":
                needEndPath = False
                pen.closePath()
            else:
                if penVerb == "moveTo":
                    if needEndPath:
                        pen.endPath()
                    needEndPath = True
                pointArgs = ((x, y) for x, y in points[startIndex:])
                getattr(pen, penVerb)(*pointArgs)
        if needEndPath:
            pen.endPath()

    def drawToPointPen(self, pen):
        self.drawToPen(SegmentToPointPen(pen))

    def text(self, txt, offset=None, font=None, fontSize=10, align=None):
        if not txt:
            return
        textStyle = TextStyle(font=font, fontSize=fontSize)
        glyphsInfo = textStyle.shape(txt)
        textStyle.alignGlyphPositions(glyphsInfo, align)
        gids = sorted(set(glyphsInfo.gids))
        paths = [textStyle.skFont.getPath(gid) for gid in gids]
        for path in paths:
            path.transform(FLIP_MATRIX)
        paths = dict(zip(gids, paths))
        x, y = (0, 0) if offset is None else offset
        for gid, pos in zip(glyphsInfo.gids, glyphsInfo.positions):
            path = paths[gid]
            self.path.addPath(path, pos[0] + x, pos[1] + y)

    def _doPathOp(self, other, operator):
        from pathops import Path, op
        path1 = Path()
        path2 = Path()
        self.drawToPen(path1.getPen())
        other.drawToPen(path2.getPen())
        result = op(
            path1,
            path2,
            operator,
            fix_winding=True,
            keep_starting_points=True,
        )
        resultPath = BezierPath()
        result.draw(resultPath)
        return resultPath

    def union(self, other):
        from pathops import PathOp
        return self._doPathOp(other, PathOp.UNION)

    def intersection(self, other):
        from pathops import PathOp
        return self._doPathOp(other, PathOp.INTERSECTION)

    def difference(self, other):
        from pathops import PathOp
        return self._doPathOp(other, PathOp.DIFFERENCE)

    def xor(self, other):
        from pathops import PathOp
        return self._doPathOp(other, PathOp.XOR)

    def removeOverlap(self):
        from pathops import Path
        path = Path()
        self.drawToPen(path.getPen())
        path.simplify(
            fix_winding=True,
            keep_starting_points=False,
        )
        resultPath = BezierPath()
        path.draw(resultPath)
        self.path = resultPath.path

    __mod__ = difference

    def __imod__(self, other):
        result = self.difference(other)
        self.path = result.path
        return self

    __or__ = union

    def __ior__(self, other):
        result = self.union(other)
        self.path = result.path
        return self

    __and__ = intersection

    def __iand__(self, other):
        result = self.intersection(other)
        self.path = result.path
        return self

    __xor__ = xor

    def __ixor__(self, other):
        result = self.xor(other)
        self.path = result.path
        return self


FLIP_MATRIX = skia.Matrix()
FLIP_MATRIX.setAffine((1, 0, 0, -1, 0, 0))


def _convertConicToCubicDirty(pt1, pt2, pt3):
    #
    # NOTE: we do a crude conversion from a conic segment to a cubic bezier,
    # for two common cases, based on the following assumptions:
    # - drawbot itself does not allow conics to be drawn
    # - skia draws conics implicitly for oval(), arc() and arcTo()
    # - for oval the conic segments span 90 degrees
    # - for arc and arcTo the conic segments do not span more than 90 degrees
    # - for arc and arcTo the conic segments are circular, never elliptical
    # For all these cases, the conic weight will be (close to) zero.
    #
    # This no longer holds once a path has been transformed with skew or x/y
    # scale, in which case we need to fall back to
    # skia.Path.ConvertConicToQuads(), but that is blocked by
    # https://github.com/kyamagu/skia-python/issues/115
    # https://github.com/justvanrossum/drawbot-skia/issues/7
    #
    (x1, y1), (x2, y2), (x3, y3) = pt1, pt2, pt3
    dx1 = x2 - x1
    dy1 = y2 - y1
    dx2 = x2 - x3
    dy2 = y2 - y3
    angle1 = math.atan2(dy1, dx1)
    angle2 = math.atan2(-dy2, -dx2)
    angleDiff = (angle1 - angle2) % (2 * math.pi)
    if angleDiff > math.pi:
        angleDiff = 2 * math.pi - angleDiff
    if abs(angleDiff - math.pi / 2) < 0.0001:
        # angle is close enough to 90 degrees, we use stupid old BEZIER_ARC_MAGIC
        handleRatio = 0.5522847498
    else:
        # Fall back to the circular assumption: |pt1 pt2| == |pt2 pt3|
        d1 = math.hypot(dx1, dy1)
        d2 = math.hypot(dx2, dy2)
        if abs(d1 - d2) > 0.00001:
            logging.warning("unsupported conic form (non-circular, non-90-degrees): conic to cubic conversion will be bad")
            # TODO: we should fall back to skia.Path.ConvertConicToQuads(),
            # but that call is currently not working.
        angleHalf = angleDiff / 2
        radius = d1 / math.tan(angleHalf)
        D = radius * (1 - math.cos(angleHalf))
        handleLength = (4 * D / 3) / math.sin(angleHalf)  # length of the bcp line
        handleRatio = handleLength / d1
    return (
        (x1 + dx1 * handleRatio, y1 + dy1 * handleRatio),
        (x3 + dx2 * handleRatio, y3 + dy2 * handleRatio),
        (x3, y3),
    )


_pathVerbsToPenMethod = {
    skia.Path.Verb.kMove_Verb: ("moveTo", 0, 1),
    skia.Path.Verb.kLine_Verb: ("lineTo", 1, 2),
    skia.Path.Verb.kCubic_Verb: ("curveTo", 1, 4),
    skia.Path.Verb.kQuad_Verb: ("qCurveTo", 1, 3),
    skia.Path.Verb.kConic_Verb: ("conicTo", 1, 3),
    skia.Path.Verb.kClose_Verb: ("closePath", 1, 1),
    # skia.Path.Verb.kDone_Verb: (None, None),  # "StopIteration", not receiving when using Python iterator
}



from fontTools.pens.recordingPen import RecordingPen, RecordingPointPen

size(200, 200)
translate(10, 10)
scale(0.39)

path = BezierPath()
path.moveTo((100, 100))
path.curveTo((200, 100), (300, 200), (300, 300))
path.lineTo((100, 300))
path.closePath()
path.oval(0, 0, 200, 90)
path.moveTo((250, 250))
path.arc((250, 250), 200, 0, 120, False)
# path.skew() will trigger bad results
# https://github.com/justvanrossum/drawbot-skia/issues/7
# path.skew(10, 20)

fill(None)
stroke(0)
strokeWidth(2)
drawPath(path)

pen = RecordingPen()
path.drawToPen(pen)

path = BezierPath()
pen.replay(path)

stroke(None)
fill(0, 0.3)
drawPath(path)

ppen = RecordingPointPen()
path.drawToPointPen(ppen)



from .drawing import Drawing
from .runner import makeDrawbotNamespace


__all__ = []
_db = Drawing()
_dbNamespace = makeDrawbotNamespace(_db)
globals().update(_dbNamespace)
__all__.extend(_dbNamespace.keys())



import os
import pathlib
import sys
import pytest
from PIL import Image
import numpy as np
from drawbot_skia.runner import makeDrawbotNamespace, runScript, runScriptSource
from drawbot_skia.drawing import Drawing


testDir = pathlib.Path(__file__).resolve().parent
apiTestsDir = testDir / "apitests"
apiTestsOutputDir = testDir / "apitests_output"
apiTestsExpectedOutputDir = testDir / "apitests_expected_output"


apiScripts = apiTestsDir.glob("*.py")


expectedFailures = [
    ("clip", "svg", "darwin"),
    ("clip", "pdf", "linux"),
    ("clip", "svg", "linux"),
    ("clip", "pdf", "win32"),
    ("clip", "svg", "win32"),
    ("fontFromPath", "pdf", "linux"),
    ("fontFromPath", "pdf", "win32"),
    ("fontFromPath", "svg", "win32"),
    ("fontFromPath2", "pdf", "linux"),
    ("fontFromPath2", "pdf", "win32"),
    ("fontVariations", "pdf", "linux"),
    ("fontVariations", "svg", "linux"),
    ("fontVariations", "pdf", "win32"),
    ("fontVariations", "svg", "win32"),
    ("image", "pdf", "linux"),
    ("image", "pdf", "win32"),
    ("imageBlendMode", "pdf", "linux"),
    ("imageBlendMode", "pdf", "win32"),
    ("language", "pdf", "linux"),
    ("language", "pdf", "win32"),
    ("pathText", "pdf", "linux"),
    ("pathText", "svg", "linux"),
    ("pathText", "pdf", "win32"),
    ("pathText", "svg", "win32"),
    ("pathTextRemoveOverlap", "pdf", "linux"),
    ("pathTextRemoveOverlap", "svg", "linux"),
    ("pathTextRemoveOverlap", "pdf", "win32"),
    ("pathTextRemoveOverlap", "svg", "win32"),
    ("textShaping", "pdf", "linux"),
    ("textShaping", "pdf", "win32"),
]


@pytest.mark.parametrize("apiTestPath", apiScripts)
@pytest.mark.parametrize("imageType", ["png", "jpg", "pdf", "svg"])
def test_apitest(apiTestPath, imageType):
    apiTestPath = pathlib.Path(apiTestPath)
    db = Drawing()
    namespace = makeDrawbotNamespace(db)
    runScript(apiTestPath, namespace)
    if not apiTestsOutputDir.exists():
        apiTestsOutputDir.mkdir()
    fileName = (apiTestPath.stem + f".{imageType}")
    outputPath = apiTestsOutputDir / fileName
    expectedOutputPath = apiTestsExpectedOutputDir / fileName
    db.saveImage(outputPath)
    if (apiTestPath.stem, imageType, sys.platform) in expectedFailures:
        # Skip late, so we can still inspect the output
        pytest.skip(f"Skipping expected failure {apiTestPath.stem}.{imageType}")
    same, reason = compareImages(outputPath, expectedOutputPath)
    assert same, f"{reason} {apiTestPath.name} {imageType}"


multipageSource = """
for i in range(3):
    newPage(200, 200)
    rect(50, 50, 100, 100)
"""


singlepageSource = """
newPage(200, 200)
rect(50, 50, 100, 100)
"""


test_data_saveImage = [
    (singlepageSource, "png", ['test.png']),
    (singlepageSource, "jpg", ['test.jpg']),
    (singlepageSource, "svg", ['test.svg']),
    (singlepageSource, "pdf", ['test.pdf']),
    (singlepageSource, "mp4", ['test.mp4']),
    (multipageSource, "png", ['test_0.png', 'test_1.png', 'test_2.png']),
    (multipageSource, "jpg", ['test_0.jpg', 'test_1.jpg', 'test_2.jpg']),
    (multipageSource, "svg", ['test_0.svg', 'test_1.svg', 'test_2.svg']),
    (multipageSource, "pdf", ['test.pdf']),
    (multipageSource, "mp4", ['test.mp4']),
]


@pytest.mark.parametrize("script, imageType, expectedFilenames", test_data_saveImage)
def test_saveImage_multipage(tmpdir, script, imageType, expectedFilenames):
    glob_pattern = f"*.{imageType}"
    tmpdir = pathlib.Path(tmpdir)
    db = Drawing()
    namespace = makeDrawbotNamespace(db)
    runScriptSource(script, "<string>", namespace)
    assert [] == sorted(tmpdir.glob(glob_pattern))
    outputPath = tmpdir / f"test.{imageType}"
    db.saveImage(outputPath)
    assert expectedFilenames == [p.name for p in sorted(tmpdir.glob(glob_pattern))]


def test_saveImage_mp4_codec(tmpdir):
    from drawbot_skia import ffmpeg
    ffmpeg.FFMPEG_PATH = ffmpeg.getPyFFmpegPath()  # Force ffmpeg from pyffmpeg
    tmpdir = pathlib.Path(tmpdir)
    db = Drawing()
    namespace = makeDrawbotNamespace(db)
    runScriptSource(multipageSource, "<string>", namespace)
    assert [] == sorted(tmpdir.glob("*.png"))
    db.saveImage(tmpdir / "test.mp4")
    db.saveImage(tmpdir / "test2.mp4", codec="mpeg4")
    expectedFilenames = ['test.mp4', 'test2.mp4']
    paths = sorted(tmpdir.glob("*.mp4"))
    assert paths[0].stat().st_size < paths[1].stat().st_size
    assert expectedFilenames == [p.name for p in paths]


def test_noFont(tmpdir):
    db = Drawing()
    # Ensure we don't get an error when font is not set
    db.text("Hallo", (0, 0))


def test_newPage_newGState():
    # Test a bug with the delegate properties of Drawing: they should
    # not return the delegate method itself, but a wrapper that calls the
    # delegate method, as the delegate object does not have a fixed
    # identity
    db = Drawing()
    fill = db.fill  # Emulate a db namespace: all methods are retrieved once
    db.newPage(50, 50)
    with db.savedState():
        fill(0.5)
        assert (255, 128, 128, 128) == db._gstate.fillPaint.color
    db.newPage(50, 50)
    fill(1)
    assert (255, 255, 255, 255) == db._gstate.fillPaint.color


def test_polygon_args():
    db = Drawing()
    db.polygon([0, 0], [0, 100], [100, 0])


def test_line_args():
    db = Drawing()
    db.line([0, 0], [0, 100])


def readbytes(path):
    with open(path, "rb") as f:
        return f.read()


def compareImages(path1, path2):
    data1 = readbytes(path1)
    data2 = readbytes(path2)
    _, ext = os.path.splitext(path1)
    if ext == ".svg":
        # Ignore line endings for svg
        data1 = data1.splitlines()
        data2 = data2.splitlines()
    if data1 == data2:
        return True, "data identical"
    if ext not in {".png", ".jpg"}:
        return False, "image data differs"
    im1 = Image.open(path1)
    im2 = Image.open(path2)
    if im1 == im2:
        return True, "images identical"
    if im1.size != im2.size:
        return False, "sizes differ"
    a1 = np.array(im1).astype(int)
    a2 = np.array(im2).astype(int)
    diff = a1 - a2
    maxDiff = max(abs(np.max(diff)), abs(np.min(diff)))
    if maxDiff < 128:
        return True, "images similar enough"
    return False, f"images differ too much, maxDiff: {maxDiff}"



import contextlib
import functools
import math
import skia
from .document import RecordingDocument
from .errors import DrawbotError
from .gstate import GraphicsState


class gstateDelegateMethod:

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, cls):
        if obj is None:
            return self

        @functools.wraps(getattr(GraphicsState, self.name))
        def wrappedMethod(*args, **kwargs):
            method = getattr(obj._gstate, self.name)
            return method(*args, **kwargs)
        return wrappedMethod


class Drawing:

    def __init__(self, document=None, flipCanvas=True):
        self._stack = []
        self._gstate = GraphicsState()
        if document is None:
            document = RecordingDocument()
        self._document = document
        self._skia_canvas = None
        self._flipCanvas = flipCanvas

    @property
    def _canvas(self):
        if self._skia_canvas is None:
            self.size(1000, 1000)  # This will create the canvas
        return self._skia_canvas

    @_canvas.setter
    def _canvas(self, canvas):
        self._skia_canvas = canvas

    def newDrawing(self):
        ...

    def endDrawing(self):
        ...

    def size(self, width, height):
        if self._document.isDrawing:
            raise DrawbotError("size() can't be called if there's already a canvas active")
        self.newPage(width,  height)

    def newPage(self, width, height):
        if self._document.isDrawing:
            self._document.endPage()
            self._gstate = GraphicsState()
        self._canvas = self._document.beginPage(width, height)
        if self._flipCanvas:
            self._canvas.translate(0, height)
            self._canvas.scale(1, -1)

    def frameDuration(self, duration):
        self._document.setFrameDuration(duration)

    def width(self):
        return self._document.pageWidth

    def height(self):
        return self._document.pageHeight

    def rect(self, x, y, w, h):
        self._drawItem(self._canvas.drawRect, (x, y, w, h))

    def oval(self, x, y, w, h):
        self._drawItem(self._canvas.drawOval, (x, y, w, h))

    def line(self, pt1, pt2):
        x1, y1 = pt1
        x2, y2 = pt2
        self._drawItem(self._canvas.drawLine, x1, y1, x2, y2)

    def polygon(self, firstPoint, *points, close=True):
        from .path import BezierPath
        bez = BezierPath()
        bez.polygon(firstPoint, *points, close=close)
        self.drawPath(bez)

    def drawPath(self, path):
        self._drawItem(self._canvas.drawPath, path.path)

    def clipPath(self, path):
        self._canvas.clipPath(path.path, doAntiAlias=True)

    # The following methods will be directly delegated to self._gstate
    fill = gstateDelegateMethod()
    stroke = gstateDelegateMethod()
    blendMode = gstateDelegateMethod()
    strokeWidth = gstateDelegateMethod()
    lineCap = gstateDelegateMethod()
    lineJoin = gstateDelegateMethod()
    lineDash = gstateDelegateMethod()
    miterLimit = gstateDelegateMethod()
    linearGradient = gstateDelegateMethod()
    radialGradient = gstateDelegateMethod()
    shadow = gstateDelegateMethod()
    font = gstateDelegateMethod()
    fontSize = gstateDelegateMethod()
    openTypeFeatures = gstateDelegateMethod()
    fontVariations = gstateDelegateMethod()
    language = gstateDelegateMethod()

    def textSize(self, txt):
        # TODO: with some smartness we can shape only once, for a
        # textSize()/text() call combination with the same text and
        # the same text parameters.
        glyphsInfo = self._gstate.textStyle.shape(txt)
        textWidth = glyphsInfo.endPos[0]
        return (textWidth, self._gstate.textStyle.skFont.getSpacing())

    def text(self, txt, position, align=None):
        if not txt:
            # Hard Skia crash otherwise
            return

        glyphsInfo = self._gstate.textStyle.shape(txt)
        blob = self._gstate.textStyle.makeTextBlob(glyphsInfo, align)

        x, y = position

        with self._savedCanvasState():
            self._canvas.translate(x, y)
            if self._flipCanvas:
                self._canvas.scale(1, -1)
            self._drawItem(self._canvas.drawTextBlob, blob, 0, 0)

    def image(self, imagePath, position, alpha=1.0):
        im = self._getImage(imagePath)
        paint = skia.Paint()
        if alpha != 1.0:
            paint.setAlpha(round(alpha * 255))
        if self._gstate.fillPaint.blendMode != "normal":
            paint.setBlendMode(self._gstate.fillPaint.skPaint.getBlendMode())
        x, y = position
        with self._savedCanvasState():
            self._canvas.translate(x, y + im.height())
            if self._flipCanvas:
                self._canvas.scale(1, -1)
            self._canvas.drawImage(im, 0, 0, paint)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def _getImage(imagePath):
        return skia.Image.open(imagePath)

    def translate(self, x, y):
        self._canvas.translate(x, y)

    def rotate(self, angle, center=(0, 0)):
        cx, cy = center
        self._canvas.rotate(angle, cx, cy)

    def scale(self, sx, sy=None, center=(0, 0)):
        if sy is None:
            sy = sx
        cx, cy = center
        if cx != 0 or cy != 0:
            self._canvas.translate(cx, cy)
            self._canvas.scale(sx, sy)
            self._canvas.translate(-cx, -cy)
        else:
            self._canvas.scale(sx, sy)

    def skew(self, sx, sy=0, center=(0, 0)):
        cx, cy = center
        if cx != 0 or cy != 0:
            self._canvas.translate(cx, cy)
            self._canvas.skew(math.radians(sx), math.radians(sy))
            self._canvas.translate(-cx, -cy)
        else:
            self._canvas.skew(math.radians(sx), math.radians(sy))

    def transform(self, matrix, center=(0, 0)):
        m = skia.Matrix()
        m.setAffine(matrix)
        cx, cy = center
        if cx != 0 or cy != 0:
            self._canvas.translate(cx, cy)
            self._canvas.concat(m)
            self._canvas.translate(-cx, -cy)
        else:
            self._canvas.concat(m)

    @contextlib.contextmanager
    def savedState(self):
        self._stack.append(self._gstate.copy())
        self._canvas.save()
        yield
        self._canvas.restore()
        self._gstate = self._stack.pop()

    @contextlib.contextmanager
    def _savedCanvasState(self):
        self._canvas.save()
        yield
        self._canvas.restore()

    def saveImage(self, fileName, **kwargs):
        if self._document.isDrawing:
            self._document.endPage()
        self._document.saveImage(fileName, **kwargs)

    # Helpers

    def _drawItem(self, canvasMethod, *items):
        shadowPaintFill, offset = self._gstate.fillPaint.skPaintShadowAndOffset
        if shadowPaintFill is not None:
            shadowPaintStroke, _ = self._gstate.strokePaint.skPaintShadowAndOffset
            dx, dy = offset
            with self._savedCanvasState():
                self._canvas.translate(dx, dy)
                if self._gstate.fillPaint.somethingToDraw:
                    canvasMethod(*items, shadowPaintFill)
                if self._gstate.strokePaint.somethingToDraw:
                    canvasMethod(*items, shadowPaintStroke)

        if self._gstate.fillPaint.somethingToDraw:
            canvasMethod(*items, self._gstate.fillPaint.skPaint)
        if self._gstate.strokePaint.somethingToDraw:
            canvasMethod(*items, self._gstate.strokePaint.skPaint)



import os
import pathlib
import pytest
import skia
from drawbot_skia.text import getShapeFuncForSkiaTypeface


testDir = pathlib.Path(__file__).resolve().parent
mutatorFontPath = testDir / "fonts" / "MutatorSans.ttf"
plexFontPath = testDir / "fonts" / "IBMPlexSansArabic-Regular.otf"


shapeTestCases = [
    (mutatorFontPath, "ABC", {}, {}, None, None, None, (
        [1, 2, 3], [0, 1, 2], [(0, 0), (400, 0), (843, 0)], (1342, 0),
     )),
    (plexFontPath, "ag0", {}, {}, None, None, None, (
        [1, 8, 56], [0, 1, 2], [(0, 0), (534, 0), (1062, 0)], (1662, 0),
     )),
    (plexFontPath, "ag0", dict(ss01=True, ss02=True, ss03=True), {}, None, None, None, (
        [2, 9, 58], [0, 1, 2], [(0, 0), (580, 0), (1160, 0)], (1760, 0),
     )),
    (plexFontPath, "أي", {}, {}, None, None, None, (
        [990, 295], [1, 0], [(0, 0), (726, 0)], (950, 0),
     )),
]


@pytest.mark.parametrize(
    "fontPath, text, features, variations, direction, language, script, expected",
    shapeTestCases,
)
def test_shape(fontPath, text, features, variations, direction, language, script, expected):
    tf = skia.Typeface.MakeFromFile(os.fspath(fontPath))
    shapeFunc = getShapeFuncForSkiaTypeface(tf)
    glyphInfo = shapeFunc(
        text,
        features=features,
        variations=variations,
        direction=direction,
        language=language,
        script=script,
    )
    expectedGids, expectedClusters, expectedPositions, expectedEndPos = expected
    assert expectedGids == glyphInfo.gids
    assert expectedClusters == glyphInfo.clusters
    assert expectedPositions == glyphInfo.positions
    assert expectedEndPos == glyphInfo.endPos



import logging
import os
import skia
from .errors import DrawbotError
from .font import makeTTFontFromSkiaTypeface, tagToInt
from .text import getShapeFuncForSkiaTypeface
from .segmenting import textSegments, reorderedSegments


class cached_property(object):

    # This exitsts in the stdlib in Python 3.8, but not earlier

    """A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property."""

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class GraphicsStateMixin:

    # Paint style

    def fill(self, *args):
        color = _colorArgs(args)
        if color is None:
            self.fillPaint = self.fillPaint.copy(somethingToDraw=False, shader=None)
        else:
            self.fillPaint = self.fillPaint.copy(color=color, somethingToDraw=True, shader=None)

    def stroke(self, *args):
        color = _colorArgs(args)
        if color is None:
            self.strokePaint = self.strokePaint.copy(somethingToDraw=False, shader=None)
        else:
            self.strokePaint = self.strokePaint.copy(color=color, somethingToDraw=True, shader=None)

    def blendMode(self, blendMode):
        if blendMode not in _blendModes:
            raise DrawbotError(f"blendMode must be one of: {_blendModesList}")
        self.fillPaint = self.fillPaint.copy(blendMode=blendMode)
        self.strokePaint = self.strokePaint.copy(blendMode=blendMode)

    def strokeWidth(self, strokeWidth):
        self.strokePaint = self.strokePaint.copy(strokeWidth=strokeWidth)

    def lineCap(self, lineCap):
        if lineCap not in _strokeCapMapping:
            raise DrawbotError(f"lineCap must be one of: {sorted(_strokeCapMapping)}")
        self.strokePaint = self.strokePaint.copy(lineCap=lineCap)

    def lineJoin(self, lineJoin):
        if lineJoin not in _strokeJoinMapping:
            raise DrawbotError(f"lineJoin must be one of: {sorted(_strokeJoinMapping)}")
        self.strokePaint = self.strokePaint.copy(lineJoin=lineJoin)

    def lineDash(self, firstValue=None, *values):
        if firstValue is None:
            if values:
                raise TypeError("lineDash() argument(s) should be None, or one or more numbers")
        if firstValue is None:
            assert not values
            self.strokePaint = self.strokePaint.copy(lineDash=None)
        else:
            self.strokePaint = self.strokePaint.copy(lineDash=(firstValue,) + values)

    def miterLimit(self, miterLimit):
        self.strokePaint = self.strokePaint.copy(miterLimit=miterLimit)

    def linearGradient(self, startPoint, endPoint, colors, locations=None):
        # MakeLinear(
        #   points: List[skia.Point],
        #   colors: List[int],
        #   positions: object = None,
        #   mode: skia.TileMode = TileMode.kClamp,
        #   flags: int = 0,
        #   localMatrix: skia.Matrix = None) -> skia.Shader
        colors = [_colorTupleToInt(_colorArgs(c)) for c in colors]
        shader = skia.GradientShader.MakeLinear(
            points=[startPoint, endPoint],
            colors=colors,
            positions=locations,
        )
        self.fillPaint = self.fillPaint.copy(shader=shader, fill=None, somethingToDraw=True)

    def radialGradient(self, startPoint, endPoint=None, colors=None, locations=None, startRadius=0, endRadius=100):
        # MakeRadial(
        #   center: skia.Point,
        #   radius: float,
        #   colors: List[int],
        #   positions: object = None,
        #   mode: skia.TileMode = TileMode.kClamp,
        #   flags: int = 0,
        #   localMatrix: skia.Matrix = None) → skia.Shader
        if startRadius != 0:
            logging.warning("radialGradient: startRadius != 0 ignored (it's not supported in drawbot-skia)")
        if endPoint is not None and endPoint != startPoint:
            logging.warning("radialGradient: endPoint argument ignored (it's not supported in drawbot-skia)")
        colors = [_colorTupleToInt(_colorArgs(c)) for c in colors]
        shader = skia.GradientShader.MakeRadial(
            center=startPoint,
            radius=endRadius,
            colors=colors,
            positions=locations,
        )
        self.fillPaint = self.fillPaint.copy(shader=shader, fill=None, somethingToDraw=True)

    def shadow(self, offset, blur=None, color=None):
        if offset is None:
            shadow = None
        else:
            shadow = (offset, blur, color)
        self.fillPaint = self.fillPaint.copy(shadow=shadow)
        self.strokePaint = self.strokePaint.copy(shadow=shadow)

    # Text style

    def font(self, fontNameOrPath, fontSize=None):
        if fontSize is not None:
            self.textStyle = self.textStyle.copy(font=fontNameOrPath, fontSize=fontSize)
        else:
            self.textStyle = self.textStyle.copy(font=fontNameOrPath)

    def fontSize(self, size):
        self.textStyle = self.textStyle.copy(fontSize=size)

    def openTypeFeatures(self, *, resetFeatures=False, **features):
        if resetFeatures:
            currentFeatures = {}
        else:
            currentFeatures = dict(self.textStyle.features)
        currentFeatures.update(features)
        self.textStyle = self.textStyle.copy(features=currentFeatures)
        return currentFeatures

    def fontVariations(self, *, resetVariations=False, **variations):
        if resetVariations:
            currentVariations = {}
        else:
            currentVariations = dict(self.textStyle.variations)
        currentVariations.update(variations)
        self.textStyle = self.textStyle.copy(variations=currentVariations)
        return currentVariations

    def language(self, language):
        self.textStyle = self.textStyle.copy(language=language)


class GraphicsState(GraphicsStateMixin):

    # The graphics state object only deals with paint and text properties,
    # other state, such as the current transformation and clip path is
    # managed by the Skia Canvas itself.

    def __init__(self, _doInitialize=True):
        if _doInitialize:
            # see self.copy()
            self.fillPaint = FillPaint()
            self.strokePaint = StrokePaint(somethingToDraw=False)
            self.textStyle = TextStyle()

    def copy(self):
        result = GraphicsState(_doInitialize=False)
        # Our main attributes are copy-on-write, so we can share them with
        # our copy
        result.fillPaint = self.fillPaint
        result.strokePaint = self.strokePaint
        result.textStyle = self.textStyle
        return result


class _ImmutableContainer:

    def __init__(self, **properties):
        self.__dict__.update(properties)
        self._names = set(properties)

    def copy(self, **properties):
        dct = {n: self.__dict__[n] for n in self._names}
        dct.update(properties)
        return self.__class__(**dct)

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if self._names != other._names:
            return False
        return all(getattr(self, n) == getattr(other, n) for n in self._names)

    def __repr__(self):
        args = ", ".join(f"{n}={self.__dict__[n]!r}" for n in sorted(self._names))
        return f"{self.__class__.__name__}({args})"


class FillPaint(_ImmutableContainer):

    somethingToDraw = True
    color = (255, 0, 0, 0)  # ARGB
    blendMode = "normal"
    shader = None
    shadow = None
    _skPaintStyle = skia.Paint.kFill_Style

    @cached_property
    def skPaint(self):
        return self._makePaint()

    def _makePaint(self):
        paint = skia.Paint(
            Color=0,
            AntiAlias=True,
            Style=self._skPaintStyle,
        )
        paint.setARGB(*self.color)
        paint.setBlendMode(_blendModeMapping[self.blendMode])
        if self.shader is not None:
            paint.setShader(self.shader)
        return paint

    @cached_property
    def skPaintShadowAndOffset(self):
        if self.shadow is None:
            return None, None
        offset, blur, color = self.shadow
        color = _colorTupleToInt(_colorArgs(color))
        if not blur:
            blur = 1
        # MaskFilter.MakeBlur(
        #   style: skia.BlurStyle,
        #   sigma: float,
        #   respectCTM: bool = True) → skia.MaskFilter
        blurMask = skia.MaskFilter.MakeBlur(skia.kNormal_BlurStyle, 0.5 * blur, False)

        paint = skia.Paint(
            Color=color,
            MaskFilter=blurMask,
            Style=self._skPaintStyle,
        )
        return paint, offset


class StrokePaint(FillPaint):

    miterLimit = 5
    strokeWidth = 1
    lineCap = "butt"
    lineJoin = "miter"
    lineDash = None
    _skPaintStyle = skia.Paint.kStroke_Style

    @cached_property
    def skPaint(self):
        paint = self._makePaint()
        paint.setStrokeMiter(self.miterLimit)
        paint.setStrokeWidth(self.strokeWidth)
        paint.setStrokeCap(_strokeCapMapping[self.lineCap])
        paint.setStrokeJoin(_strokeJoinMapping[self.lineJoin])
        if self.lineDash is not None:
            intervals = self.lineDash
            if len(intervals) % 2:
                # Skia requires the intervals list to be of even length;
                # doubling the list matches macOS/CoreGraphics behavior
                intervals = intervals * 2
            paint.setPathEffect(skia.DashPathEffect.Make(intervals, 0))
        return paint


_strokeCapMapping = dict(
    butt=skia.Paint.Cap.kButt_Cap,
    round=skia.Paint.Cap.kRound_Cap,
    square=skia.Paint.Cap.kSquare_Cap,
)

_strokeJoinMapping = dict(
    miter=skia.Paint.Join.kMiter_Join,
    round=skia.Paint.Join.kRound_Join,
    bevel=skia.Paint.Join.kBevel_Join,
)

# NOTE: 'plusDarker' is missing
_blendModeMapping = {
    "softLight": skia.BlendMode.kSoftLight,  # lighten or darken, depending on source
    "destinationOut": skia.BlendMode.kDstOut,  # destination trimmed outside source
    "clear": skia.BlendMode.kClear,  # replaces destination with zero: fully transparent
    "sourceIn": skia.BlendMode.kSrcIn,  # source trimmed inside destination
    "destinationOver": skia.BlendMode.kDstOver,  # destination over source
    "hardLight": skia.BlendMode.kHardLight,  # multiply or screen, depending on source
    # "???": skia.BlendMode.kDst,  # preserves destination
    "xOR": skia.BlendMode.kXor,  # each of source and destination trimmed outside the other
    "hue": skia.BlendMode.kHue,  # hue of source with saturation and luminosity of destination
    "screen": skia.BlendMode.kScreen,  # multiply inverse of pixels, inverting result; brightens destination
    # "???": skia.BlendMode.kLastMode,  # last valid value
    "difference": skia.BlendMode.kDifference,  # subtract darker from lighter with higher contrast
    "overlay": skia.BlendMode.kOverlay,  # multiply or screen, depending on destination
    # "???": skia.BlendMode.kModulate,  # product of premultiplied colors; darkens destination
    "colorBurn": skia.BlendMode.kColorBurn,  # darken destination to reflect source
    # "???": skia.BlendMode.kSrc,  # replaces destination
    "plusLighter": skia.BlendMode.kPlus,  # sum of colors
    "destinationIn": skia.BlendMode.kDstIn,  # destination trimmed by source
    "destinationAtop": skia.BlendMode.kDstATop,  # destination inside source blended with source
    "saturation": skia.BlendMode.kSaturation,  # saturation of source with hue and luminosity of destination
    # "???": skia.BlendMode.kLastSeparableMode,  # last blend mode operating separately on components
    "sourceAtop": skia.BlendMode.kSrcATop,  # source inside destination blended with destination
    "sourceOut": skia.BlendMode.kSrcOut,  # source trimmed outside destination
    # "???": skia.BlendMode.kLastCoeffMode,  # last porter duff blend mode
    "normal": skia.BlendMode.kSrcOver,  # source over destination
    "copy": skia.BlendMode.kSrcOver,  # source over destination
    "colorDodge": skia.BlendMode.kColorDodge,  # brighten destination to reflect source
    "darken": skia.BlendMode.kDarken,  # darker of source and destination
    "luminosity": skia.BlendMode.kLuminosity,  # luminosity of source with hue and saturation of destination
    "multiply": skia.BlendMode.kMultiply,  # multiply source with destination, darkening image
    "lighten": skia.BlendMode.kLighten,  # lighter of source and destination
    "color": skia.BlendMode.kColor,  # hue and saturation of source with luminosity of destination
    "exclusion": skia.BlendMode.kExclusion,  # subtract darker from lighter with lower contrast
}


class TextStyle(_ImmutableContainer):

    fontSize = 10
    features = {}  # won't get mutated
    variations = {}  # won't get mutated
    language = None
    font = None

    def __init__(self, **properties):
        super().__init__(**properties)

    @cached_property
    def skFont(self):
        typeface, ttFont = self._getTypefaceAndTTFont(self.font)
        if self.variations and "fvar" in ttFont:
            typeface = self._cloneTypeface(typeface, ttFont, self.variations)
        font = self._makeFontFromTypeface(typeface, self.fontSize)
        return font

    @property
    def ttFont(self):
        _, ttFont = self._getTypefaceAndTTFont(self.font)
        return ttFont

    @staticmethod
    def _getTypefaceAndTTFont(fontNameOrPath):
        cacheKey = fontNameOrPath
        if cacheKey not in _fontCache:
            if fontNameOrPath is None:
                typeface = skia.Typeface(None)
            else:
                fontNameOrPath = os.fspath(fontNameOrPath)
                if not os.path.exists(fontNameOrPath):
                    typeface = skia.Typeface(fontNameOrPath)
                else:
                    typeface = skia.Typeface.MakeFromFile(fontNameOrPath)
                    if typeface is None:
                        raise DrawbotError(f"can't load font: {fontNameOrPath}")
            ttFont = makeTTFontFromSkiaTypeface(typeface)
            _fontCache[cacheKey] = typeface, ttFont
        return _fontCache[cacheKey]

    @staticmethod
    def _makeFontFromTypeface(typeface, size):
        font = skia.Font(typeface, size)
        font.setForceAutoHinting(False)
        font.setHinting(skia.FontHinting.kNone)
        font.setSubpixel(True)
        font.setEdging(skia.Font.Edging.kAntiAlias)
        return font

    @staticmethod
    def _cloneTypeface(typeface, ttFont, variations):
        fvar = ttFont.get("fvar")
        defaultLocation = {a.axisTag: variations.get(a.axisTag, a.defaultValue) for a in fvar.axes}
        tags = [a.axisTag for a in fvar.axes]
        location = [(tag, variations.get(tag, defaultLocation[tag])) for tag in tags]
        makeCoord = skia.FontArguments.VariationPosition.Coordinate
        rawCoords = [makeCoord(tagToInt(tag), value) for tag, value in location]
        coords = skia.FontArguments.VariationPosition.Coordinates(rawCoords)
        pos = skia.FontArguments.VariationPosition(coords)
        fontArgs = skia.FontArguments()
        fontArgs.setVariationDesignPosition(pos)
        return typeface.makeClone(fontArgs)

    @cached_property
    def _shapeFunc(self):
        return getShapeFuncForSkiaTypeface(self.skFont.getTypeface())

    def shape(self, txt):
        segments, baseLevel = textSegments(txt)
        segments = reorderedSegments(segments, baseLevel)
        startPos = (0, 0)
        glyphsInfo = None
        for runChars, script, bidiLevel, index in segments:
            runInfo = self._shapeFunc(
                runChars,
                fontSize=self.skFont.getSize(),
                startPos=startPos,
                startCluster=index,
                flippedCanvas=True,
                features=self.features,
                variations=self.variations,
                language=self.language,
            )
            if glyphsInfo is None:
                glyphsInfo = runInfo
            else:
                glyphsInfo.gids += runInfo.gids
                glyphsInfo.clusters += runInfo.clusters
                glyphsInfo.positions += runInfo.positions
                glyphsInfo.endPos = runInfo.endPos
            startPos = runInfo.endPos
        glyphsInfo.baseLevel = baseLevel
        return glyphsInfo

    def alignGlyphPositions(self, glyphsInfo, align):
        textWidth = glyphsInfo.endPos[0]
        if align is None:
            align = "left" if not glyphsInfo.baseLevel else "right"
        xOffset = 0
        if align == "right":
            xOffset = -textWidth
        elif align == "center":
            xOffset = -textWidth / 2
        glyphsInfo.positions = [(x + xOffset, y) for x, y in glyphsInfo.positions]

    def makeTextBlob(self, glyphsInfo, align):
        self.alignGlyphPositions(glyphsInfo, align)
        builder = skia.TextBlobBuilder()
        builder.allocRunPos(self.skFont, glyphsInfo.gids, glyphsInfo.positions)
        return builder.make()


# Font cache dict
# - keys: font name or path string
# - values: (skTypeface, ttFont) tuples
_fontCache = {}


def clearFontCache():
    _fontCache.clear()


def _colorTupleToInt(color):
    intColor = 0
    for channel, shift in zip(color, (24, 16, 8, 0)):
        intColor |= channel << shift
    return intColor


def _colorArgs(args):
    """Convert drawbot-style fill/stroke arguments to a tuple containing
    ARGB int values."""
    if not args:
        return None
    alpha = 1
    if len(args) == 1:
        if args[0] is None:
            return None
        r = g = b = args[0]
    elif len(args) == 2:
        r = g = b = args[0]
        alpha = args[1]
    elif len(args) == 3:
        r, g, b = args
    elif len(args) == 4:
        r, g, b, alpha = args
    else:
        assert 0
    return tuple(min(255, max(0, round(v * 255))) for v in (alpha, r, g, b))


_blendModesList = [
    'normal',
    'multiply',
    'screen',
    'overlay',
    'darken',
    'lighten',
    'colorDodge',
    'colorBurn',
    'softLight',
    'hardLight',
    'difference',
    'exclusion',
    'hue',
    'saturation',
    'color',
    'luminosity',
    'clear',
    'copy',
    'sourceIn',
    'sourceOut',
    'sourceAtop',
    'destinationOver',
    'destinationIn',
    'destinationOut',
    'destinationAtop',
    'xOR',
    'plusDarker',
    'plusLighter',
]

_blendModes = set(_blendModesList)



values = [
    # sw, lj, lc, dash
    (5, "miter", "butt", [5]),
    (12, "miter", "butt", [5, 10]),
    (8, "round", "round", [8, 10, 15]),
    (5, "round", "round", [0, 7, 7, 7]),
    (5, "round", "round", [None]),  # reset
    (5, "round", "round", []),  # reset
]

size(100, 100 * len(values))

stroke(0)
fill(None)

for sw, lj, lc, dash in values:
    strokeWidth(sw)
    lineJoin(lj)
    lineCap(lc)
    lineDash(*dash)
    polygon((18, 7), (10, 90), (80, 80), (60, 10), close=False)
    translate(0, 100)



size(200, 200)

bez = BezierPath()
bez.moveTo((20, 20))
bez.lineTo((20, 100))
bez.lineTo((100, 100))
bez.lineTo((100, 180))
bez.curveTo((150, 180), (180, 150), (180, 100))
bez.lineTo((180, 50))
bez.qCurveTo((180, 20), (150, 20))

fill(1, 0, 0)
stroke(0)
strokeWidth(10)
drawPath(bez)

bez.closePath()

fill(None)
stroke(1)
translate(40, 15)
scale(0.7)
lineCap("round")
lineJoin("round")

drawPath(bez)



import pathlib
import subprocess


testDir = pathlib.Path(__file__).resolve().parent
testScript = testDir / "apitests" / "oval.py"


def test_runner_app(tmpdir):
    outputPath = pathlib.Path(tmpdir / "test.png")
    assert not outputPath.exists()
    # assert 0, outputPath
    args = ["drawbot", testScript, outputPath]
    subprocess.check_output(args)
    assert outputPath.exists()



import os
import pathlib
import skia
from drawbot_skia.font import makeTTFontFromSkiaTypeface
from drawbot_skia.gstate import GraphicsState


testDir = pathlib.Path(__file__).resolve().parent
fontPath = testDir / "fonts" / "MutatorSans.ttf"


def test_font():
    skTypeface = skia.Typeface.MakeFromFile(os.fspath(fontPath))
    ttf = makeTTFontFromSkiaTypeface(skTypeface)
    assert ttf["name"].getName(6, 3, 1).toUnicode() == "MutatorMathTest-LightCondensed"
    fvar = ttf["fvar"]
    assert fvar.axes[0].axisTag == "wdth"
    assert ["name", "fvar"] == list(ttf.tables.keys())


def test_font_gs():
    gs = GraphicsState()
    gs.font(fontPath)
    assert "fvar" in gs.textStyle.ttFont



from setuptools import setup, find_packages


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="drawbot-skia",
    use_scm_version={"write_to": "src/drawbot_skia/_version.py"},
    description="Drawbot implements a simple drawing API to generate 2D vector graphics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Just van Rossum",
    author_email="justvanrossum@gmail.com",
    url="http://github.com/justvanrossum/drawbot-skia",
    license="Apache License 2.0",
    entry_points={
        'console_scripts': ['drawbot=drawbot_skia.__main__:main'],
    },
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=[
        "skia-python",
        "fonttools[unicode]",
        "numpy",  # unlisted skia-python dependency, TODO: is this true?
        "uharfbuzz",
        "python-bidi",
        "unicodedata2",
    ],
    setup_requires=["setuptools_scm"],
    extras_require={
        "mp4": ["pyffmpeg==1.6.1"],  # pin for now, this package is not very stable
    },
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Graphics",
    ],
)



size(600, 500)

fill(1)
rect(0, 0, width(), height())
fill(0)
font("../fonts/IBMPlexSansArabic-Regular.otf")
fontSize(100)
openTypeFeatures(ss01=True, ss02=True, ss03=True)
openTypeFeatures(resetFeatures=True, ss03=True)
text("Aag0", (50, 50))
w, h = textSize("Aag0")
text("---", (50 + w, 50))

openTypeFeatures(resetFeatures=True)
fontSize(60)
s = "بلديهما abcde المتحدة"
text(s, (width() - 50, 150))

text("Alfabet", (width() / 2, 220))
text("البشريةً", (width() / 2, 290))

font("Helvetica")  # simply test loading an installed font



size(200, 200)

path = BezierPath()
path.moveTo((100, 100))
path.arc((100, 100), 80, 230, 170, False)
path.arc((100, 100), 60, 30, 120, True)
path.closePath()

stroke(0)
fill(None)
strokeWidth(4)
drawPath(path)

# arc(center, radius, startAngle, endAngle, clockwise)
# Arc with center and a given radius, from startAngle to endAngle, going clockwise if clockwise is True and counter clockwise if clockwise is False.

# arcTo(point1, point2, radius)
# Arc defined by a circle inscribed inside the angle specified by three points: the current point, point1, and point2. The arc is drawn between the two points of the circle that are tangent to the two legs of the angle.
